# Copyright 2025 hingebase

# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at

#     http://www.apache.org/licenses/LICENSE-2.0

# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

__all__ = ["run"]

import contextlib
from typing import TYPE_CHECKING

import filelock

from . import _server

if TYPE_CHECKING:
    from _typeshed import StrPath


def run(root: StrPath) -> None:
    with (
        contextlib.chdir(root),
        filelock.FileLock("mahoraga.lock", timeout=0, thread_local=False),
    ):
        _server.run()
